var
  long  cog

pub start(Qpin, APDpin, LPOK, Thresh_p, mute_p, tx_lock, tx_head_p, tx_buffer_p, bias_p)
  ''this starts the pulse delay measureing code in it's own cog.
  ''Qpin = this is the I/O pin number of the Q-switch signal pulse
  ''APDpin = this is the I/O pin number of the APD that detects the transmitted laser pulse
  ''LPOK = this is the I/O pin number of the LPOK logic line to the clock/accumulator boards
  ''Thresh_p = this is a pointer to a hub memory location containing the current delay threshold below wich LPOK is pulsed
  ''    (limited to 8-bit unsigned values, {thresh + bias} = maximum P to P clocks where LPOK is pulsed)
  ''mute_p = this is a pointer to a logic value in hub memory.  If mute is TRUE data is nolonger passed on to the serial cog
  ''tx_lock = this is the lock number of the lock protecting the serial transmit buffer
  ''tx_head_p = this is a pointer to the head counter of the serial transmit buffer
  ''tx_buffer_p = this is a pointer to the start of the transmit buffer.  This is a 16 element BYTE buffer
  ''bias_p = this is a pointer to a value that is subtracted off of the pulse to pulse time sent over the serial port.
  ''    (bias_p is a 32-bit number that is set to N*2^7 where N is an 8-bit unsigned integer)
  ''    the measured delay has a minimum value of 14 clocks
  

  stop
  'move passed variables into ASM code
  Qpin_mask := 1 << Qpin
  APDpin_mask := 1 << APDpin
  xLPOK := LPOK
  xThresh_p := Thresh_p
  xmute_p := mute_p
  xtx_lock := tx_lock
  xtx_head_p := tx_head_p
  xtx_buffer_p := tx_buffer_p
  xbias_p := bias_p

  'start up the ASM code
  result := cog := cognew(@delay_entry,0)+1

pub stop
  ''stops the pulse delay measuring code if it alredy is running

  if cog
    cogstop(cog~ - 1)

dat

                org     0
                'setup CTRA as a 1-shot to generate LPOK pulses
delay_entry     mov     frqa,           #1                      'count up by one
                or      CMA,            xLPOK                   'combine the pin with the counter mode
                mov     ctra,           CMA                     'actavate the CTRA with the proper mode on the right pin
                'set DIRA to output for the pins that are outputs
                mov     xtemp1,          #1                     'prepare a pin mask
                shl     xtemp1,          xLPOK                  'shift the mask bit into the righ position
                or      dira,           xtemp1                  'combine the pin mask with DIRA
                'load up variables
                rdlong  xThresh_n,      xThresh_p               'read in my initial good/bad threshold
                rdlong  xbias_n,        xbias_p                 'read in my initial bias value
                add     xThresh_n,      xbias_n                 'add the bias value to the LPOK threshold.  
                
                '################ start of core loop
:loop           waitpeq Qpin_mask,      Qpin_mask               'wait for the rising edge of Q-switch pulse
                mov     C1,             cnt                     'store the value of CNT
                mov     phsa,           #0                      'keep my pulse generator from timing out
                waitpeq APDpin_mask,    APDpin_mask             'wait for the rising edge of APD pulse
                mov     C2,             cnt                     'store the value of CNT
                sub     C2,             C1                      'subtract the starting time "C1" from the ending time "C2."  "C2" is now the pulse width in clocks
                cmp     C2,             xThresh_n       wc      'set C flag if C2 is less than xThresh_n
        if_c    mov     phsa,           CMA_width               'have CTRA generate a pulse of 40 clock cycles (code is not time critical from this point on)
                rdlong  xThresh_n,      xThresh_p               'read in an updated Thresh value
                rdlong  xbias_n,        xbias_p                 'read in an updated bias value
                add     xThresh_n,      xbias_n                 'add the bias value to the LPOK threshold.
                'write out results 
                rdlong  xtemp1,         xmute_p                 'read the mute flag
                tjnz    xtemp1,         #:end_loop              'if the mute flag is TRUE skip all the rest of this junk
                mov     timeout,        cnt                     'mark the current CNT so we can timeout if the tx_buffer of the serial port is busy for too long       
:ck_lock        mov     xtemp1,         cnt                     'catch the current CNT
                sub     xtemp1,         timeout                 'subtract my initial CNT from the current CNT resulting in the time elapsed being stored in "xtemp1"
                cmp     wait_time,      xtemp1          wc      'is the time elapsed greater than my wait_time limit?
        if_c    jmp     #:end_loop                              'out of time, go to the end of the loop        
                lockset xtx_lock                        wc      'check if the TX lock is free and set the lock if it is free
        if_c    jmp     #:ck_lock                               'keep waiting till the lock is free
                        'write the least significant byte of C2 to the serial X-mit buffer
                rdlong  xtemp1,         xtx_head_p              'read in the serial buffer head counter
                add     xtemp1,         xtx_buffer_p            'adjust the pointer to the serial transmit byte buffer by the value of the head counter
                sub     C2,             xbias_n         wc,wz   'subtract off the bias and update the flags
        if_be   mov     C2,             #1                      'if the result of the bias subtraction is zero or negative make the result equal one
                max     C2,             #255                    'limit the result to a maximum of 255
                wrbyte  C2,             xtemp1                  'write the least significant byte of C2 into the transmit buffer at the HEAD location        
                        'update the buffer control variable to signal the serial cog it has new data
                rdlong  xtemp1,         xtx_head_p              'read in the value of the head counter
                add     xtemp1,         #1                      'add one
                and     xtemp1,         #$F                     'limit the values to 0-16
                wrlong  xtemp1,         xtx_head_p              'write back the updated value of the head counter        
                lockclr xtx_lock                                'free up the serial port X-mit lock  
:end_loop       jmp     #:loop                                  'jump back to the start of the loop                                      




'initialised data and parameters passed into the ASM code
Qpin_mask       long    0-0
APDpin_mask     long    0-0
xLPOK           long    0-0
xThresh_p       long    0-0
xmute_p         long    0-0
xtx_lock        long    0-0
xtx_head_p      long    0-0
xtx_buffer_p    long    0-0
CMA             long    %0_00100_00000000000000000000_000000    'Counter Mode ctrA
CMA_width       long    -40                             '"mov phsa, CMA_width" to make a pulse
wait_time       long    11000                                   'maximum number of clock cycles to wait for the serial port to be free 
xbias_p         long    0-0             

                fit     496
'any RES statements go below here
xtemp1          res     1
xtemp2          res     1
xThresh_n       res     1
C1              res     1
C2              res     1
timeout         res     1
xbias_n         res     1