con
  cntMin = 400                  'Minimum waitcnt value to prevent lock-up
  Max_uS = 2350                 'longest valid servo pulse (in microseconds)
  Min_uS = 300                  'shortest valid servo pulse (in microseconds)
  Precision = 3                 'how small of a step the search uses
  Start_offset = 150            'how far to offset form the assumed position to start
  PP_wait_div = 150             '1 / PP_wait_div seconds between search pulses
  Num_ch = 7                    'number of servos (physical_channels - 1)

obj
  i2c : "basic_i2c_driver.spin" 'a simple I2C bus driver for updating the EEPROM
  
var                             
  'these variables are unique to each instance of this object.
  long servo_lock               'hardware lock to protect the software servo locks
  long EE_lock                  'lock to arbitrate access to the I2C boot EEPROM                

pub Init | temp, index, pin 
{{initalise all the timing for this monster code.  
Init needs to be called at startup and anytime the clock speed changes
Only needs to be called once for all instances of the object.}}
   us := clkfreq / 1_000_000                  ' Clock cycles for 1 us
   servo_lock := locknew                      'check out a lock for servo arbitration
   EE_lock := locknew                         'check out a lock for eeprom arbitration

pri res_ch(channel):old | temp
  'wait for successful set of servo_lock
  repeat
    temp := lockset(servo_lock)
  while temp
  'run software lock for servo channels
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  old := byte[temp][3]                                  'store old soft lock byte
  byte[temp][3] := true                                 'set soft lock byte
  'free up servo_lock
  lockclr(servo_lock)

pri free_ch(channel):old | temp
  'wait for successful set of servo_lock
  repeat
    temp := lockset(servo_lock)
  while temp
  'run software lock for servo channels
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  old := byte[temp][3]                                  'store old soft lock byte
  byte[temp][3] := false                                'clear soft lock byte
  'free up servo_lock
  lockclr(servo_lock)


pub Num_ch_con : a_number
  ''return the internal constant Num_ch 'oops! i can do this directly, this code is redundant
  a_number := Num_ch
                   

Pub Pulseout_uS(Pin, Duration) | SyncCNT
''generate a pulse on Pin, Duration microseconds long.
  Duration := Min_uS #> Duration <# Max_uS     'limit Duration to valid values                                                     
  dira[pin]~~                                  ' Set to output                                         
  SyncCNT := cnt + 1000                        'snag value of cnt + some time
  waitcnt(SyncCNT)                             'sync to cnt
  !outa[pin]                                   'set Pin to opposite state                                 
  {do math while we wait for CNT to reach our target}                
  SyncCNT += ((Duration * us) #> cntMin)       'modify sync point to a future time. duration*(clk cycles for us)
  waitcnt(SyncCNT)                             'wait until clk gets there                             
  !outa[pin]                                   'return Pin to orig. state

con
  Offset_up_err = 35            'offset error of up search method
  error_limit = 1

pub Find_pos_up(Pin,AssumedPos):RealPos | error
{{search for the current position of the servo.
assumes signal pin is Pin and mode pin is Pin+1}}
  dira[Pin+1] := 0                 'make the mode pin an input
  error := 0                    'counter for how many time's search has happened
  RealPos := AssumedPos - Start_offset - Offset_up_err    'start the search just below AssumedPos
  {setup for main loop}
  Pulseout_uS(Pin,RealPos)      'pulse the servo
  waitcnt(clkfreq/PP_wait_div + cnt)    'wait for 5ms
  if ina[Pin+1]                 'if servo's internal position matches the command
    RealPos += Offset_up_err    'adjust for offset error
    return                      'return RealPos
  {main loop}
  repeat                        'loop through the avalable positions
    RealPos += Precision                'increment tested position by 5us
    if RealPos => Max_uS        'check if the command pulse is too long
      RealPos := Min_uS         'start over at the min pulse length
      error++
    Pulseout_uS(Pin,RealPos)    'pulse the servo
    waitcnt(clkfreq/PP_wait_div + cnt)  'wait for 5ms
  until ina[Pin+1] or (error > error_limit)    'I'm stuck in here till I match positions with the servo's internals or I give up
  waitcnt(clkfreq/50 +cnt)      'kill a glitch
  if error > error_limit                  'did I give up?
    RealPos := -1               'something's not right.  return an error code.
  RealPos += Offset_up_err      'adjust position for offset error

con
  Offset_down_err = 25            'offset error of down search method

pub Find_pos_down(Pin,AssumedPos):RealPos | error
{{search for the current position of the servo.
assumes signal pin is Pin and mode pin is Pin+1
for some reson this is more glitchy than the up search}}
  dira[Pin+1]~                  'make the mode pin an input
  error := 0                    'counter for how many time's search has happened
  RealPos := AssumedPos + Start_offset - Offset_down_err    'start the search just abovve AssumedPos
  {setup for main loop}
  Pulseout_uS(Pin,RealPos)      'pulse the servo
  waitcnt(clkfreq/PP_wait_div + cnt)    'wait for 5ms
  if ina[Pin+1]                 'if servo's internal position matches the command
    RealPos += Offset_down_err    'adjust for offset error
    return                      'return RealPos
  {main loop}
  repeat                        'loop through the avalable positions
    RealPos -= Precision                'decrement tested position by 5us
    if RealPos =< Min_uS        'check if the command pulse is too short
      RealPos := Max_uS         'start over at the max pulse length
      error++
    Pulseout_uS(Pin,RealPos)    'pulse the servo
    waitcnt(clkfreq/PP_wait_div + cnt)  'wait for 5ms
  until ina[Pin+1] or (error > 3)    'I'm stuck in here till I match positions with the servo's internals or I give up
  waitcnt(clkfreq/50 +cnt)      'kill a glitch
  if error > 3                  'did I give up?
    RealPos := -1               'something's not right.  return an error code.
  RealPos += Offset_down_err      'adjust position for offset error

con
  Radius_from_target = 20       'maximum number of counts u can miss a target position

Pub Move_servo(Channel, Target_pos):temp2 | signal_pin, Lo_limit, Hi_limit, speed, temp, real_pos, old_real_pos, old_error_code, error_code
{{decode Target_pos (0 to 255) into the appropriate pulse for Channel.
Move the servo to the above position at Speed_of_movement
Verify that the servo reached Target_pos
Set an object golbal variable with the appropriate error_code
Also returns the error_code if the function is run inline}}

  'clear error code to indicate that i'm active
  error_code := 0

  'wait for Channel to be free
  repeat
    temp := res_ch(Channel)
  while temp

  'extract important data from the central storage area
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  signal_pin := byte[temp]                              'extract the pin # I'm using
  speed := long[temp][1]                                'extract how fast I should move (uS/loop)
  Lo_limit := long[temp][2]                             'extract one of my motion limits (uS)
  Hi_limit := long[temp][3]                             'extract the other motion limit (uS)
  old_real_pos := long[temp][4]                         'extract my prevously found position (uS)
  old_error_code := long[temp][5]                       'extract my last error code (not using it at the moment)
  
  'catch a startup glitch
  if old_real_pos == 0                                  'if this is the first run
    old_real_pos := find_pos_up(signal_pin,1200)        'find my current position and assign it to old_real_pos

  'scale Target_pos to this axis
  Target_pos := 0 #> Target_pos <# 255                  'limit range
  Target_pos := (Hi_limit - Lo_limit) * Target_pos / 255 + Lo_limit             're-scale

  'activate the mode pin
  outa[signal_pin+1] := 1   'enable servo drive with my special circuit
  dira[signal_pin+1] := 1   'enable servo drive with my special circuit

  'move to Target_pos in steps of Speed
  repeat temp from old_real_pos to Target_pos step speed
    Pulseout_us(signal_pin,temp)
    waitcnt(clkfreq/60 + cnt)
  repeat 8                      'make shure i've reached the target position
    pulseout_us(signal_pin,Target_pos)
    waitcnt(clkfreq/60 + cnt)  

  'deactivate the mode pin
  dira[signal_pin+1] := 0   'disable servo drive with my special circuit
  
  'Search for current position and update error_code
  real_pos := Find_pos_up(signal_pin,Target_pos)        'find current position
  'real_pos := target_pos                                'temporarily disable position confirmation
  temp := ||(Target_pos - real_pos)                     'how far off am I?
  if temp =< Radius_from_target                 'everything's fine
    error_code += 1
  if temp > Radius_from_target                  'short of my goal
    error_code += 8
  if real_pos < 0               'search for my current position failed
    error_code += 2
  temp := ||(old_real_pos - real_pos)                   'some math
  if temp =< Radius_from_target                 'I didn't even move!
    error_code += 4

  'update history variables in the servo infor array
  temp := servo_dat_point(Channel)  'pre calculate the pointer and store it in Temp                                            
  long[temp][4] := Target_pos 'real_pos         'update the old position
  long[temp][5] := error_code       'update the old error code

  'return the latest error code
  temp2 := error_code

  'free up Channel
  free_ch(Channel)


Pub servo_dat_point(Channel):pointer                    'return the pointer into the servo array for Channel
  Channel := 0 #> Channel <# Num_ch                     'limit Channel to a valid servo channel 
  pointer := (Channel * 6 * 4) + @servo_dat             'add in the appropriate offset (6 longs per record by 4 bytes per long)
  

Pub search_servo(Channel): Real_Pos | temp, signal_pin
  {{search for the current position of the servo on Channel
  returns time in units of 10uS}}

  'wait for Channel to be free
  repeat
    temp := res_ch(Channel)
  while temp

  'extract important data from the central storage area
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  signal_pin := byte[temp]                              'extract the pin # I'm using

  Real_pos := Find_pos_up(signal_pin, 1320) / 10        'find my real position scaled to units of 10uS 

  'free up Channel
  free_ch(Channel)    
  
Pub force_move(Channel, Pos):Final_pos | temp, signal_pin
  {{Force the specified servo channel to move to a raw position
  and return what position the servo ended up in}}

  'wait for Channel to be free
  repeat
    temp := res_ch(channel)     
  while temp

  'extract important data from the central storage area
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  signal_pin := byte[temp]                              'extract the pin # I'm using

  'move the servo
  outa[signal_pin+1] := 1   'enable servo drive with my special circuit
  dira[signal_pin+1] := 1   'enable servo drive with my special circuit

  repeat 30
    Pulseout_us(signal_pin,Pos*10)
    waitcnt(clkfreq/60 + cnt)

  dira[signal_pin+1] := 0   'disable servo drive with my special circuit

  'verify the servo moved as commanded.
  Final_pos := Find_pos_up(Signal_pin, Pos*10) / 10     'find my real Position and scale to units of 10us

  'update history variables in the centeral storage area
  temp := servo_dat_point(Channel)  'pre calculate the pointer and store it in Temp                                            
  long[temp][4] := Pos * 10 'real_pos         'update the old position

  'free up Channel
  free_ch(Channel)


Pub Set_low_limit(Channel, limit):status | temp
  {{set Low_limit for Channel in the servo information structure}}
  limit := limit * 10                                   'scale up to 1us units                        
  limit := Min_uS #> limit <# Max_uS                    'limit Limit to valid values

  'wait for Channel to be free
  repeat
    temp := res_ch(channel)     
  while temp

  'update the servo infor sturcture                                           
  long[servo_dat_point(Channel)][2] := limit            'update the old Low_limit value

  'free up Channel
  free_ch(Channel)

  'return a status code.  (not needed at the moment)
  status := 0
                                              
    
Pub Set_High_limit(Channel, limit):status | temp
  {{set High_limit for Channel in the servo information structure}}
  limit := limit * 10                                   'scale up to 1us units                        
  limit := Min_uS #> limit <# Max_uS                    'limit Limit to valid values

  'wait for Channel to be free
  repeat
    temp := res_ch(channel)     
  while temp

  'update the servo infor sturcture                                           
  long[servo_dat_point(Channel)][3] := limit            'update the old High_limit value

  'free up Channel
  free_ch(Channel)

  'return a status code.  (not needed at the moment)
  status := 0


Pub Set_speed(Channel, speed):status | temp
  {{set the Speed variable of Channel in the servo information structure}}
  speed := 1 #> speed <# 80     'limit to sane speeds only

  'wait for Channel to be free
  repeat
    temp := res_ch(channel)     
  while temp

  'update the servo info sturcture                                           
  long[servo_dat_point(Channel)][1] := speed            'update the old High_limit value

  'free up Channel
  free_ch(Channel)

  'return a status code.  (not needed at the moment)
  status := 0

Pub Save_servo_data | temp, i, temp2
  {{this function updates the boot eeprom so that the
  limits and speeds of each channel match what the controller is currently using}}

  'wait for the eeprom to be free
  repeat
    temp := lockset(EE_lock)
  while temp

  'write Ch0 to Num_ch
  repeat i from 0 to Num_ch
    temp := servo_dat_point(i)                          'pointer to the start of Chi's entry
    temp += 4                                           'skip the pin long
    i2c.WritePage(i2c#BootPin, i2c#EEPROM, temp, temp, 12) 'write out the next 3 longs of data
    'wait for the write to finish
    temp2 := cnt ' prepare to check for a timeout
    repeat while i2c.WriteWait(i2c#BootPin, i2c#EEPROM, temp)                'wait for the write to finish
      if cnt - temp2 > clkfreq / 10
        quit ' waited more than a 1/10 second for the write to finish

  'free up the eeprom
  lockclr(EE_lock)  


Dat
{{ single structure containing all the setup and history information
for each servo axis.  shared by ALL of the instances of this object}}

us            long      0-0                             'store timing variable in common for ALL instances of this object
{                       base pin,speed,low limit, high limit, last position, not used}
servo_dat     long      0, 15, 1580, 1060, 0-0, 0-0    'current data for 8 axis of servos.  (no real limit actually)
              long      2, 15, 1130, 1650, 0-0, 0-0
              long      4, 15, 1200, 1790, 0-0, 0-0
              long      6, 13, 1000, 1500, 0-0, 0-0
              long      8, 15, 1000, 1500, 0-0, 0-0
              long      10, 15, 1000, 1500, 0-0, 0-0
              long      12, 13, 1000, 1500, 0-0, 0-0
              long      14, 15, 1000, 1500, 0-0, 0-0